unit GLTextures;

{
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
interface
}

interface

uses
 SysUtils, FileMaps, DelphiGL;

function LoadTexture(FileName:string;var dx,dy:single):integer;

implementation

Function PowerOf2(Target:integer):integer;
 begin
  Result:=1;
  while Result<Target do Result:=Result shl 1;
 end;

type
 TBMP256=packed record
  bfType:array[1..2] of char;
  bfSize:LongInt;
  bfReserved:LongInt;
  bfOffBits:LongInt;
 //---
  biSize:LongInt;
  biWidth:LongInt;
  biHeight:LongInt;
  biPlanes:word;
  biBitCount:word;
  biCompression:longint;
  biSizeImage:longint;
  biXPelsPerMeter:longint;
  biYPelsPerMeter:longint;
  biClrUsed:Longint;
  biClrImportant:longint;
 end;

 TPCX256=packed record
  manufacturer:byte;
  version     :byte;
  encoding    :byte; // 1
  BitsPerPixel:byte; // 8
  xmin,ymin,xmax,ymax:word; // ClientRect
//  hres,vres:word;
//  palette:array[0..47] of byte;
//  reserved:byte;
//  planes:byte;
//  BytesPerLine:word;
//  PaletteType:word;
//  dummy:array[0..57] of byte;
 end;

 TBMPPalette=packed array[#0..#255] of record r,g,b,a:byte end;
 TPCXPalette=packed array[ 0.. 255] of record b,g,r  :byte end;

 TPixels =array[word] of integer;
 PPixels =^TPixels;

function LoadTexture(FileName:string;var dx,dy:single):integer;
 var
  Mapping:TFileMapping;
  Pixels:PPixels;
  width,height:integer;

 procedure SetSize(AWidth,AHeight:integer);
  begin
   Width :=PowerOf2(AWidth);
   Height:=PowerOf2(AHeight);
   dx:=AWidth /Width;
   dy:=AHeight/Height;
   GetMem(Pixels,Width*Height*4);
  end;

 procedure LoadBMP;
 var
  bmp:^TBMP256;
  pal:^TBMPPalette;
  pix:pchar;
  x,y:integer;
  i,j:integer;
  k  :char;
  cl :integer;
  w4 :integer;
 begin
  bmp:=Mapping.Base;
  SetSize(bmp.biWidth,bmp.biHeight);
  case bmp.biBitCount of
   8 : begin
    pal:=Mapping.Data(14+bmp.biSize);
    pix:=Mapping.Data(bmp.bfOffBits);
    i:=0;
    j:=Width*bmp.biHeight;
    for y:=0 to bmp.biHeight-1 do begin
     dec(j,Width);
     for x:=0 to bmp.biWidth-1 do begin
      k:=pix[i+x];
      with pal[k] do cl:=r shl 16+g shl 8+b;
      Pixels[j+x]:=cl;
     end;
     inc(i,bmp.biWidth);
    end;
   end;
   24 : begin
    pix:=Mapping.Data(bmp.bfOffBits);
    i:=0;
    j:=0;
    w4:=3*bmp.biWidth+bmp.biWidth and 1;
    for y:=0 to bmp.biHeight-1 do begin
     //dec(j,Width);
     for x:=0 to bmp.biWidth-1 do begin
      cl:=$ff000000+ord(pix[i+3*x]) shl 16+ord(pix[i+3*x+1]) shl 8+ord(pix[i+3*x+2]);
      Pixels[j+x]:=cl;
     end;
     inc(i,w4);
     inc(j,Width);
    end;
   end;
   else raise Exception.Create('unsupported bitmap bpp : '+IntToStr(bmp.biBitCount));
  end;
 end;

 procedure LoadPCX;
 var
  pcx:^TPCX256;
  pal:^TPCXPalette;
  pix:pchar;
  x,y:integer;
  i,j:integer;
  k,l:integer;
  cl:integer;
 begin
  pcx:=Mapping.Base;
  SetSize(pcx.xmax-pcx.xmin+1,pcx.ymax-pcx.ymin+1);
  pix:=Mapping.Data(128);
  pal:=Mapping.Data(Mapping.Size-sizeof(TPCXPalette));
  i:=0; // pixel index
  j:=0; // source index
  l:=0; // repeat count
  cl:=0; // to avoid a compilater warning...
  for y:=pcx.ymin to pcx.ymax do begin
   for x:=pcx.xmin to pcx.xmax do begin
    if l=0 then begin // need a new pixel
     k:=ord(pix[j]); inc(j);
     if (k and $C0)=$C0 then begin // repeat code
      l:=k and $3F; // repeat count
      k:=ord(pix[j]); inc(j); // get pixel
     end else begin
      l:=1; // only once
     end;
     with pal[k] do cl:=r shl 16+g shl 8+b;
    end;
    Pixels[i+x]:=cl;
    dec(l);
   end;
   inc(i,Width);
  end;
 end;

 begin
  Mapping:=TFileMapping.Create(FileName);
  try
   case word(Mapping.Base^) of
    $4D42 : LoadBMP;
    $050A : LoadPCX;
    else raise exception.Create('unknow texture file format');
   end;
   glEnable(GL_TEXTURE_2D);
   glGenTextures(1,@Result);
   glBindTexture(GL_TEXTURE_2D, Result);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
   glTexImage2D(GL_TEXTURE_2D, 0, 4, Width, Height,0, GL_RGBA, GL_UNSIGNED_BYTE, Pixels);
   FreeMem(Pixels);
  finally
   Mapping.Free;
  end;
 end;

end.
