unit LoadCR2;
{
  Poser Figure Loader (C)2001 by Paul TOTH <tothpaul@free.fr>
  http://tothpaul.free.fr
}
{
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
interface
}

interface

{$define log}
{-$define logchar}

uses
 Windows,Classes,SysUtils,GLRender,LoadOBJ;

Type
 TCR2Reader=class(TGLReader)
 private
  Figure:TextFile;
  line  :integer;
  LastChar:char;
  Wavefronts:TStringList;
  procedure Load;
  function ReadChar:char;
  procedure Blanks;
  procedure LineFeed;
  procedure Drop(s:string);
  function Skip(c:char):boolean;
  procedure Ignore;
  function ReadToken:string;
  function ReadInteger:integer;
  procedure ReadSection;
  procedure SkipSection;
  procedure ReadWavefront;
  procedure ReadActor;
  procedure ReadActorChannels(Actor:string);
  procedure ReadActorChannelsTargetGeom(Actor:string);
  procedure ReadActorChannelsTargetGeomDeltas;
 public
  Constructor Create(AFileName:string);
  Destructor Destroy; override;
 end;

implementation

Constructor TCR2Reader.Create(AFileName:string);
 begin
  Wavefronts:=TStringList.Create;
  Groups:=TStringList.Create;
  with TStringList(Groups) do begin
   Sorted:=True;
   Duplicates:=dupIgnore;
  end;
  AssignFile(Figure,AFileName);
  Reset(Figure);
  if ioResult<>0 then raise exception.create('unabled to open file');
  Load;
 end;

Destructor TCR2Reader.Destroy;
 var
  i:integer;
 begin
  Closefile(Figure);
  for i:=0 to Wavefronts.count-1 do TOBJReader(Wavefronts.Objects[i]).Free;
  Wavefronts.Free;
  inherited;
 end;

Function TCR2Reader.ReadChar:char;
 begin
  if eof(Figure) then result:=#27 else
  if eoln(Figure) then result:=#0 else begin
   read(Figure,Result);
  {$ifdef log}{$ifdef logchar}write(result);{$endif}{$endif}
  end;
 end;

procedure TCR2Reader.Blanks;
 begin
  while LastChar in [#9,' '] do LastChar:=ReadChar;
 end;

procedure TCR2Reader.LineFeed;
 begin
  while not (LastChar in [#10,#27]) do LastChar:=ReadChar;
  while LastChar=#10 do begin
   inc(Line);
   LastChar:=ReadChar;
   Blanks;
  end;
 end;

Procedure TCR2Reader.Drop(s:string);
 var
  i:integer;
 begin
  Blanks;
  for i:=1 to length(s) do begin
   if LastChar<>s[i] then raise Exception.Create(s+' expected at line '+IntToStr(line));
   LastChar:=ReadChar;
  end;
 end;

function TCR2Reader.Skip(c:char):boolean;
 begin
  Result:=(LastChar=c);
  if Result then LastChar:=ReadChar;
 end;

procedure TCR2Reader.Ignore;
 begin
  if lastChar='{' then SkipSection else LineFeed;
 end;

Procedure TCR2Reader.Load;
 begin
  Line:=1;
  LastChar:=ReadChar;
  Drop('{'); LineFeed;
  while not Skip('}') do ReadSection;
 end;

procedure TCR2Reader.ReadSection;
 var
  s:string;
 begin
  s:=ReadToken;
  if s='figureResFile' then ReadWavefront else
  if s='actor' then ReadActor else Ignore;
 end;

function TCR2Reader.ReadToken:string;
 begin
  Blanks;
  Result:='';
  while upcase(LastChar) in ['A'..'Z','0'..'9'] do begin
   Result:=Result+LastChar;
   LastChar:=ReadChar;
  end;
 end;

function TCR2Reader.ReadInteger:integer;
 begin
  Blanks;
  Result:=0;
  while upcase(LastChar) in ['0'..'9'] do begin
   Result:=10*Result+ord(LastChar) and 15;
   LastChar:=ReadChar;
  end;
 end;

procedure TCR2Reader.SkipSection;
 var
  count:integer;
 begin
  Drop('{');
  LineFeed;
  count:=1;
  repeat
   if Skip('{') then inc(count) else
   if Skip('}') then dec(count) else LastChar:=ReadChar;
  until count=0;
  LineFeed;
 end;

procedure TCR2Reader.ReadWavefront;
 var
  s:string;
  i:integer;
 begin
  blanks;
  while LastChar<>#10 do begin
   if LastChar=':' then s:='' else s:=s+LastChar;
   LastChar:=ReadChar;
  end;
  LineFeed;
  for i:=0 to Wavefronts.count-1 do if Wavefronts[i]=s then exit;
  Wavefronts.AddObject(s,TOBJReader.Create(s));
 end;

procedure TCR2Reader.ReadActor;
 var
  name,figure,s:string;
 begin
  name:=ReadToken;
  Drop(':');
  figure:=ReadToken;
// {$ifdef log}writeln('actor ',name,':',figure);{$endif}
  LineFeed;
  Drop('{');
  LineFeed;
  while LastChar<>'}' do begin
   s:=ReadToken;
   if s='channels' then ReadActorChannels(name) else Ignore;
  end;
  Drop('}');
  LineFeed;
 end;

procedure TCR2Reader.ReadActorChannels(Actor:string);
 var
  s:string;
 begin
  LineFeed;
  Drop('{');
  LineFeed;
  while LastChar<>'}' do begin
   s:=ReadToken;
   if s='targetGeom' then ReadActorChannelsTargetGeom(Actor) else Ignore;
  end;
  Drop('}');
  LineFeed;
 end;

procedure TCR2Reader.ReadActorChannelsTargetGeom(Actor:string);
 var
  target:string;
  s:string;
 begin
  target:=ReadToken;
 {$ifdef log}write(Actor,'.targetGeom ',target);{$endif}
  LineFeed;
  Drop('{');
  LineFeed;
  while LastChar<>'}' do begin
   s:=ReadToken;
   if s='indexes' then begin ReadActorChannelsTargetGeomDeltas; end else Ignore;
  end;
  Drop('}');
  LineFeed;
 {$ifdef log}writeln;{$endif}
 end;

procedure TCR2Reader.ReadActorChannelsTargetGeomDeltas;
 var
  count:integer;
  s:string;
  i:integer;
 begin
  count:=ReadInteger;
  LineFeed; Drop('numbDeltas');
  LineFeed; Drop('deltas');
  LineFeed; Drop('{');
  LineFeed;
  i:=0;
  repeat
//   writeln(ReadToken);
   Ignore;
   inc(i);
  until LastChar='}';
  Drop('}');
  LineFeed;
  {$ifdef log}write(' ',i,'/',count,' deltas');{$endif}
 end;

{$ifdef log}
Initialization
 AllocConsole;
{$endif}
end.
