unit LoadMD2;
// Quake II MD2 Loader (c)2002 by Paul TOTH <tothpaul@free.fr>

{
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
interface
}

interface

uses
 Classes,SysUtils,GLRender;

Type
 TMD2Reader=class(TGLReader)
 private
  Stream:TStream;
 public
  Constructor Create(AFileName:string);
  Destructor Destroy; override;
  procedure NextFrame; override;
 end;

implementation

Type
 TMD2Header=record
  Ident      :integer;
  Version    :integer;

  SkinWidth  :integer;
  SkinHeight :integer;
  FrameSize  :integer;

  SkinCount  :integer;
  VertexCount:integer;
  MapCount   :integer;
  FaceCount  :integer;
  GLCmdCount :integer;
  FrameCount :integer;

  SkinOfs    :integer;
  MapOfs     :integer;
  FaceOfs    :integer;
  FrameOfs   :integer;
  GLCmdOfs   :integer;
  EndOfFile  :integer;
 end;

 TMD2Face=record
  a,b,c:smallint;
  ta,tb,tc:smallint;
 end;
 TMD2Faces=array[0..4095] of TMD2Face;

 TMD2Mapping=record u,v:smallint end;
 TMD2Mappings=array[0..2047] of TMD2Mapping;

 TMD2Vertex=record
  x,y,z:byte;
  light:byte;
 end;

 TMD2Frame=record
  Scale:record x,y,z:single end;
  Trans:record x,y,z:single end;
  Name :array[0..15] of char;
 end;
 TMD2FrameVertices=array[0..2047] of TMD2Vertex;

Constructor TMD2Reader.Create(AFileName:string);
 var
  head :TMD2Header;
  mface:array of TMD2Face;
  maps :array of TMD2Mapping;
  i:integer;
 begin
  Stream:=TFileStream.Create(AFileName,fmOpenRead or fmShareDenyNone);
  Stream.ReadBuffer(head,SizeOf(head));
  FrameCount:=head.FrameCount;
  SetLength(Vertices,head.VertexCount);
  SetLength(Faces,head.FaceCount);
  SetLength(Mappings,head.FaceCount);

  SetLength(mface,head.FaceCount);
  Stream.Position:=head.FaceOfs;
  Stream.ReadBuffer(mface[0],head.FaceCount*sizeof(TMD2Face));

  for i:=0 to head.FaceCount-1 do begin
   Faces[i].a:=mface[i].c;
   Faces[i].b:=mface[i].b;
   Faces[i].c:=mface[i].a;
  end;

  SetLength(maps,head.MapCount);
  Stream.Position:=head.MapOfs;
  Stream.ReadBuffer(maps[0],head.MapCount*SizeOf(TMD2Mapping));
  for i:=0 to head.FaceCount-1 do begin
   Mappings[i,0].u:=maps[mface[i].tc].u/head.skinwidth;
   Mappings[i,0].v:=maps[mface[i].tc].v/head.skinheight;
   Mappings[i,1].u:=maps[mface[i].tb].u/head.skinwidth;
   Mappings[i,1].v:=maps[mface[i].tb].v/head.skinheight;
   Mappings[i,2].u:=maps[mface[i].ta].u/head.skinwidth;
   Mappings[i,2].v:=maps[mface[i].ta].v/head.skinheight;
  end;
  maps:=nil;

  mface:=nil;

  Stream.Position:=head.FrameOfs;
  Groups:=TStringList.Create;

  SetLength(Colors,1);
  Colors[0].r:=1;
  Colors[0].g:=1;
  Colors[0].b:=1;

  SetLength(Materials,1);
  Materials[0].Name:='Skin';
  Materials[0].TwoSide:=False;
  Materials[0].Texture:=0;

  SetLength(Textures,1);
  Textures[0]:=TTexture.Create;
  Textures[0].FileName:='';

  Transparent:=false;

  NextFrame;
 end;

procedure TMD2Reader.NextFrame;
 var
  count:integer;
  frame:TMD2Frame;
  verts:array of TMD2Vertex;
  i:integer;
 begin
  Stream.ReadBuffer(Frame,Sizeof(Frame));
  Groups.Text:=Frame.Name;

  count:=Length(Vertices);

  SetLength(verts,count);
  Stream.ReadBuffer(verts[0],Length(Vertices)*SizeOf(TMD2Vertex));

  Vertices:=nil; // unalloc previous Vertices
  SetLength(Vertices,Count); // alloc new ones

  for i:=0 to count-1 do begin
   Vertices[i].x:=Frame.Scale.x*Verts[i].x+Frame.Trans.x;
   Vertices[i].y:=Frame.Scale.y*Verts[i].y+Frame.Trans.y;
   Vertices[i].z:=Frame.Scale.z*Verts[i].z+Frame.Trans.z;
  end;

  verts:=nil;
 end;

Destructor TMD2Reader.Destroy;
 begin
  Stream.Free;
  Groups.Free;
  inherited;
 end;

end.
