unit PCX;

// Load/Save 256 colors PCX (c)2001 by Paul TOTH <tothpaul@free.fr>
{
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
interface
}

interface

procedure Register;

implementation

uses
 Windows, Classes,Graphics, SysUtils;

procedure Register;
 begin
  { all is done un initialization part }
 end;

resourcestring
 SInvalidPCX = 'Image PCX incorrecte';

procedure InvalidPCX;
 begin
  raise EInvalidGraphic.CreateRes(@SInvalidPCX);
 end;

Type
 TPCX=packed record
  manufacturer:byte; // 10
  version     :byte; // 5
  encoding    :byte; // 1
  BitsPerPixel:byte; // 8
  xmin,ymin,xmax,ymax:word; // ClientRect
  hres,vres:word;
  palette:array[0..47] of byte;
  reserved:byte;
  planes:byte;
  BytesPerLine:word;
  PaletteType:word;
  dummy:array[0..57] of byte;
 end;

 TPCXImage=class(TBitmap)
  private
  // to bad, TBitmap doesn't virtualize this :(
   procedure ReadStream(Stream: TStream; Size: Longint);
   procedure WriteStream(Stream: TStream; WriteSize: Boolean);
  protected
  // DFM streaming
   procedure ReadData(Stream:TStream); override;
   procedure WriteData(Stream:TStream); override;
  public
  // file streaming
   procedure LoadFromStream(Stream: TStream); override;
   procedure SaveToStream(Stream: TStream); override;
 end;

// <TBitmap_Like>
// TBitmap ReadStream & WriteStream aren't virtual :(
procedure TPCXImage.LoadFromStream(Stream: TStream);
begin
  ReadStream(Stream, Stream.Size - Stream.Position);
end;

procedure TPCXImage.SaveToStream(Stream: TStream);
begin
  WriteStream(Stream, False);
end;

procedure TPCXImage.ReadData(Stream: TStream);
var
  Size: Longint;
begin
  Stream.Read(Size, SizeOf(Size));
  ReadStream(Stream, Size);
end;

procedure TPCXImage.WriteData(Stream: TStream);
begin
  WriteStream(Stream, True);
end;
// </TBitmap_Like>

{$define usepal}

procedure TPCXImage.ReadStream(Stream: TStream; Size:Longint);
var
 PCX:TPCX;
 PIX:PChar;
 RLE:PChar;
 w,h,ww:integer;
 pcxpal:packed record
  tag256:byte;
  colors:array[0..255,0..2] of byte;
 end;
 bmp:packed record
  bmiHeader:TBitmapInfoHeader;
  bmiColors:array[0..255] of {$ifdef usepal}word{$else}TRGBQuad{$endif};
 end;
 BitsInfo:TBitmapInfo absolute bmp;
 x,y,i,j,k,l:integer;
{$ifdef usepal}
 winpal:TMaxLogPalette;
 logpal:TLogPalette absolute winpal;
{$endif}
begin
 if Size<SizeOf(PCX)+SizeOf(pcxpal) then InvalidPCX;
 Stream.ReadBuffer(PCX,SizeOf(PCX));

 if (PCX.Manufacturer<>10)
 or (PCX.Version<>5)
 or (PCX.Encoding<>1)
 or (PCX.BitsPerPixel<>8) then InvalidPCX;

 w:=pcx.xmax-pcx.xmin+1;
 h:=pcx.ymax-pcx.ymin+1;
 ww:=(w+3) and (not 3);

 GetMem(PIX,ww*h);
 try
  i:=ww*h; // pixel index
  k:=0; // color index
  l:=0; // repeat count
  dec(size,SizeOf(PCX)+sizeof(pcxpal));
  GetMem(RLE,size);
  try
   Stream.ReadBuffer(RLE^,size);
   j:=0; // RLE Index
   for y:=0 to h-1 do begin
    dec(i,ww);
    for x:=0 to w-1 do begin
     if l=0 then begin // need a new pixel
      k:=ord(RLE[j]); inc(j);
      if (k and $C0)=$C0 then begin // repeat code
       l:=k and $3F; // repeat count
       k:=ord(RLE[j]); inc(j);
      end else begin
       l:=1; // only once
      end;
     end;
     PIX[i+x]:=chr(k);
     dec(l);
    end;
   end;
  finally
   Freemem(RLE);
  end;

  if j<>Size then InvalidPCX;

  Stream.ReadBuffer(pcxpal,sizeof(pcxpal));
  if pcxpal.tag256<>$0C then InvalidPCX;

 {$ifdef usepal}
  PixelFormat:=pf8Bit;
  winpal.palVersion:=$300;
  winpal.palNumEntries:=256;
  for i:=0 to 255 do begin
   winpal.palPalEntry[i].peRed  :=pcxpal.Colors[i,0];
   winpal.palPalEntry[i].peGreen:=pcxpal.Colors[i,1];
   winpal.palPalEntry[i].peBlue :=pcxpal.Colors[i,2];
   winpal.palPalEntry[i].peFlags:=PC_NOCOLLAPSE;
  end;
  Palette:=CreatePalette(logpal);
 {$endif}
  Width:=w;
  Height:=h;

  FillChar(bmp,sizeof(bmp),0);
  with bmp do begin
   with bmiHeader do begin
    biSize    :=SizeOf(bmp.bmiHeader);
    biWidth   :=w;
    biHeight  :=h;
    biPlanes  :=1;
    biBitCount:=8;
    biClrUsed :=256;
    biClrImportant:=256;
   end;
   for i:=0 to 255 do begin
   {$ifdef usepal}
    bmiColors[i]:=i;
   {$else}
    bmiColors[i].rgbRed     :=pcxpal.colors[i,0];
    bmiColors[i].rgbGreen   :=pcxpal.colors[i,1];
    bmiColors[i].rgbBlue    :=pcxpal.colors[i,2];
    bmiColors[i].rgbReserved:=0;
   {$endif}
   end;
  end;
  SetDiBits(Canvas.Handle,Handle,0,h,PIX,BitsInfo,{$ifdef usepal}DIB_PAL_COLORS{$else}DIB_RGB_COLORS{$endif});

 finally
  FreeMem(PIX);
 end;
end;

procedure TPCXImage.WriteStream(Stream: TStream; WriteSize: Boolean);
var
 PCX:TPCX;
 PIX:PChar;
 ww:integer;
 bmp:packed record
  bmiHeader:TBitmapInfoHeader;
  bmiColors:array[0..255] of {$ifdef usepal}word{$else}TRGBQuad{$endif};
 end;
 BitsInfo:TBitmapInfo absolute bmp;
 RLE:string;
 size:integer;
 x,y,c:integer;
 value:byte;
 i,j:integer;
 pcxpal:packed record
  tag256:byte;
  colors:array[0..255,0..2] of byte;
 end;
{$ifdef usepal}
 winpal:array[0..255] of TPaletteEntry;
{$endif}
 procedure add(b:byte);
  begin
   inc(j);
   if j>Length(RLE) then SetLength(RLE,Length(RLE)+ww);
   RLE[j]:=chr(b);
  end;

 procedure flush;
  begin
   if c=0 then exit;
   if c=1 then begin
    if (value and $C0)=$C0 then add($C1);
   end else begin
    add($C0+c);
   end;
   add(value);
  end;

begin
 fillchar(pcx,sizeof(pcx),0);
 pcx.manufacturer:=10;
 pcx.version:=5;
 pcx.encoding:=1;
 pcx.BitsPerPixel:=8;
 pcx.xmin:=0;
 pcx.ymin:=0;
 pcx.xmax:=Width-1;
 pcx.ymax:=Height-1;
 pcx.hres:=Width;
 pcx.vres:=Height;
 pcx.planes:=1;
 pcx.BytesPerLine:=Width;
 pcx.PaletteType:=1;
 size:=sizeof(pcx);

 ww:=(Width+3) and (not 3);
 GetMem(PIX,ww*Height);
 try
  FillChar(bmp,sizeof(bmp),0);
  with bmp do begin
   with bmiHeader do begin
    biSize    :=SizeOf(bmp.bmiHeader);
    biWidth   :=width;
    biHeight  :=height;
    biPlanes  :=1;
    biBitCount:=8;
    biClrUsed :=256;
    biClrImportant:=256;
   end;
  end;
  GetDiBits(Canvas.Handle,Handle,0,Height,PIX,BitsInfo,{$ifdef usepal}DIB_PAL_COLORS{$else}DIB_RGB_COLORS{$endif});
 {$ifdef usepal}
  GetPaletteEntries(Palette,0,256,winpal);
 {$endif}

  SetLength(RLE,ww*Height);
  j:=0;
  i:=ww*Height;
  for y:=0 to Height-1 do begin
   c:=0;
   dec(i,ww);
   for x:=0 to Width-1 do begin
    if ord(PIX[i+x])=value then begin
     if c<$3F then inc(c) else begin
      add($ff); // $C0+$3F = $FF
      add(value);
      c:=1;
     end;
    end else begin
     flush;
     value:=ord(PIX[i+x]);
     c:=1;
    end;
   end;
   flush;
  end;
  inc(size,j);
  fillchar(pcxpal,sizeof(pcxpal),0);
  pcxpal.tag256:=$0C;
  for i:=0 to 255 do begin
  {$ifdef usepal}
   pcxpal.colors[i,0]:=winpal[i].peRed;
   pcxpal.colors[i,1]:=winpal[i].peGreen;
   pcxpal.colors[i,2]:=winpal[i].peBlue;
  {$else}
   pcxpal.colors[i,0]:=bmp.bmiColors[i].rgbRed;
   pcxpal.colors[i,1]:=bmp.bmiColors[i].rgbGreen;
   pcxpal.colors[i,2]:=bmp.bmiColors[i].rgbBlue;
  {$endif}
  end;
  inc(size,sizeof(pcxpal));
  if WriteSize then Stream.WriteBuffer(Size,SizeOf(Size));
  Stream.WriteBuffer(pcx,sizeof(pcx));
  Stream.WriteBuffer(rle[1],j);
  Stream.WriteBuffer(pcxpal,sizeof(pcxpal));
 finally
  FreeMem(PIX);
 end;
end;

initialization
 TPicture.RegisterFileFormat('pcx', 'Zsoft paintbrush', TPCXImage);
finalization
 TPicture.UnregisterGraphicClass(TPCXImage);

end.
