unit TEX;

// Load/Save 24/32 bits Textures files (c)2001 by Paul TOTH <tothpaul@free.fr>
{
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
interface
}

interface

uses
 Windows, Classes,Graphics, SysUtils;

Type
 TTEXImage=class(TBitmap)
  private
  // to bad, TBitmap doesn't virtualize this :(
   procedure ReadStream(Stream: TStream; Size: Longint);
   procedure WriteStream(Stream: TStream; WriteSize: Boolean);
  protected
  // DFM streaming
   procedure ReadData(Stream:TStream); override;
   procedure WriteData(Stream:TStream); override;
  public
  // file streaming
   procedure LoadFromStream(Stream: TStream); override;
   procedure SaveToStream(Stream: TStream); override;
 end;


implementation

resourcestring
 SInvalidTEX = 'Image TEX invalide';

procedure InvalidTEX;
 begin
  raise EInvalidGraphic.CreateRes(@SInvalidTEX);
 end;

Type
 TTEX=packed record
  sign:word;
  depth:word;
  width:integer;
  height:integer;
 end;

// <TBitmap_Like>
// TBitmap ReadStream & WriteStream aren't virtual :(
procedure TTEXImage.LoadFromStream(Stream: TStream);
begin
  ReadStream(Stream, Stream.Size - Stream.Position);
end;

procedure TTEXImage.SaveToStream(Stream: TStream);
begin
  WriteStream(Stream, False);
end;

procedure TTEXImage.ReadData(Stream: TStream);
var
  Size: Longint;
begin
  Stream.Read(Size, SizeOf(Size));
  ReadStream(Stream, Size);
end;

procedure TTEXImage.WriteData(Stream: TStream);
begin
  WriteStream(Stream, True);
end;
// </TBitmap_Like>

procedure RGB2BGR(var p);
 var
  i:array[0..2] of byte absolute p;
  t:byte;
 begin
  t:=i[2]; i[2]:=i[0]; i[0]:=t;
 end;

procedure TTEXImage.ReadStream(Stream: TStream; Size:Longint);
var
 TEX:TTEX;
 p:pchar;
 x,y:integer;
begin
 if Size<SizeOf(TEX) then InvalidTEX;
 Stream.ReadBuffer(TEX,SizeOf(TEX));
 if (TEX.sign<>$DD3D) then InvalidTEX;

 case TEX.depth of
  3: PixelFormat:=pf24bit;
  4: PixelFormat:=pf32bit;
  else InvalidTEX;
 end;

 if Size<>(TEX.Width*TEX.Height*TEX.Depth+SizeOf(TEX)) then InvalidTEX;

 Width :=TEX.Width;
 Height:=TEX.Height;
 for y:=0 to TEX.Height-1 do begin
  p:=ScanLine[y];
  Stream.ReadBuffer(p^,TEX.Width*TEX.Depth);
  for x:=0 to TEX.width-1 do begin
   RGB2BGR(p[0]);
   inc(p,TEX.Depth);
  end;
 end;
end;

Function PowerOf2(Target:integer):integer;
 begin
  Result:=1;
  while Result<Target do Result:=Result shl 1;
 end;

procedure TTEXImage.WriteStream(Stream: TStream; WriteSize: Boolean);
var
 TEX:TTEX;
 BMP:TBitmap;
 REC:TRect;
 x,y:integer;
 p1,p2:PChar;
begin
 TEX.sign:=$DD3D;
 if PixelFormat=pf32bit then
  TEX.Depth:=4
 else begin
  TEX.Depth:=3;
 end;
 TEX.Width :=PowerOf2(Width);
 TEX.Height:=PowerOf2(Height);
 BMP:=TBitmap.Create;
 BMP.Width :=TEX.Width;
 BMP.Height:=TEX.Height;
 if TEX.Depth=3 then BMP.PixelFormat:=pf24bit else BMP.PixelFormat:=pf32Bit;
 REC:=Rect(0,0,TEX.Width,TEX.Height);
 BMP.Canvas.StretchDraw(REC,Self);

 if WriteSize then begin
  x:=SizeOf(TEX)+TEX.Width*TEX.Height*TEX.Depth;
  Stream.WriteBuffer(x,SizeOf(x));
 end;
 Stream.WriteBuffer(TEX,SizeOf(TEX));
 for y:=0 to TEX.Height-1 do begin
  p1:=BMP.ScanLine[y];
  p2:=p1;
  for x:=0 to TEX.width-1 do begin
   RGB2BGR(p2[0]);
   inc(p2,TEX.Depth);
  end;
  Stream.WriteBuffer(p1^,TEX.Width*TEX.Depth);
 end;
 BMP.Free;
end;

initialization
 TPicture.RegisterFileFormat('tex', 'MySoft Texture File', TTEXImage);
finalization
 TPicture.UnregisterGraphicClass(TTEXImage);

end.
